#!/usr/bin/python3
from threading import Thread, RLock, Timer
import importlib
import time
import datetime
import socket
import pa_rtp

# ----------------------------------------------------------------------------
# Configuration
# ----------------------------------------------------------------------------

# import network config
#
# This configuration is independent from the individual client/server within
# one OVF network. It has to be changed only if multiple independent OVF
# networks shall be operated within one ip network.
net_cfg = importlib.import_module("config_net")

# Time interval in s for periodic PTT-ON messages if PTT is pressed
#
# This value must be shorter than timeout_ptt in the server.
time_ptt = 20

# Timeout for minimum receiving after LISTEN event
timeout_recv = 60

# ----------------------------------------------------------------------------

verbose = True

class OVF_Client:
    """ OVF-Client independent from user interface (UI)

    The RTP audio streams are provided by Pulseaudio and encapsulated in the
    module pa_rtp.py.

    Receiver State
    ==============

    The receivers main_state is either 'idle' (the speaker is muted), 'recv',
    or 'send'. Additionally, there are the boolean states state_listen and
    state_int.

    'idle'
        This is the inital state of the client. State_listen is always false.

    'recv'
        The state 'recv' is modified by state_listen: if false, main_state
        'idle' is entered after a timout; if true, state 'idle' never is
        entered and all audio traffic can be heared.

    'send'
        Usually, on enter, exit, and on a regular basis via timer_ptt() the
        server is informed by PTT-ON/PTT-OFF messages. Hoever, main_state
        'send' is modified by state_int: If true, no PTT-ON/PTT-OFF events are
        sent to the server, hence the QSO stays internal.

    This is the state diagram:

    idle            | recv                     | send                  |
    ================+==========================+=======================+=======
    RTP-tx mute     | RTP-tx mute              | RTP-tx unmute         |
    RTP-rx mute     | RTP-rx unmute            | RTP-rx mute           | state_
    ================+==========================+=======================+
                    |                state_int = false                 | listen
                    +==========================+=======================+=======
                  <-+-timer_recv()             | timer_ptt()           | false
                    |                          |                       |
                    |                          | on enter: PTT-ON      |
                    |                          | on exit:  PTT-OFF     |
                    +--------------------------+ - - - - - - - - - - - +-------
                    |                          |                       |
                    |                          |                       | true
                    |                          |                       |
    - - - - - - - - +==========================+=======================+=======
    enter: LISTEN-INT-REQ            state_int = true
    - - - - - - - - +==========================+=======================+=======
                  <-+-timer_recv()             | enter: LISTEN-INT-REQ | false
                    |                          |                       |
                    +--------------------------+ - - - - - - - - - - - +-------
                    |                          |                       | true
                    |                          |                       |
    ----------------+--------------------------+-----------------------+-------

    The state (state_main, state_listen, state_int) may only be accessed by
    the following, thread-safe event methods:

        User Events
        -----------
        ui_ev_PTT_press()   ->  state_main = 'send'
        ui_ev_PTT_release() ->  state_main = 'recv'
        ui_ev_press_int()   ->  toggle state_int;
                                state_main = 'recv' indirect via
                                             LISTEN-INT-REQ -> LISTEN-INT
        ui_ev_press_listen()->  toggle 'idle'/'recv'
                                set state_listen false ('idle'),
                                                 true ('recv'),
                                                 toggle ('send')
        Network Events
        --------------
        ev_LISTEN()         ->  state_main = 'recv'; state_int = False
        ev_LISTEN_INT()     ->  state_main = 'recv'; state_int = True
        ev_LISTEN_INT_REJ() ->  ev_LISTEN()

        Internal Events
        ---------------
        ev_recv_timeout()   ->  state_main = 'idle'; state_int = False
        ev_update_ui()      ->  read-only, last call of all events above


    User Interface UI
    =================

    The user interface is updated by ev_update_ui().
    All above mentioned events update the user interface.
    Additionally, the user interface is updated by:

        ui_change_volume(int)
        server_lost()
        server_found()

    All indicators of the user interface are controlled by the callback
    function 

        ui_display_callback(RX, TX, Listen, Int, Con, Vol).

    RX, TX, Listen, Int, Con are booleans controlling the LED state.
    Vol is the volume to display as integer in [0..100].


    List of Threads
    ===============

    -------------------------------------------------------------------------
    tcp_listen      in: LISTEN                      ev_LISTEN()
                        LISTEN-INT                  ev_LISTEN_INT()
                        OVF-REGISTER-OK             server_found()
    -------------------------------------------------------------------------
    keep_connected  not server_is_connected()       out: OVF-REGISTER
    -------------------------------------------------------------------------
    timer           timout_recv                     ev_recv_timeout
                    timout_ptt                      out: PTT-ON
    -------------------------------------------------------------------------


    Connection to TRX-Server
    ========================

    The server is searched automatically in keep_connected(). Whenever the
    connection state changes, the UI is informed via the ui_display_callback.

    The clients looks for the server on udp with an OVF-REGISTER broadcast.
    The server answers over tcp with OVF-REGISTER-OK. The client checks the
    presence of the server periodically via tcp with PING. If the server gets
    a PING, the sender is also added to known_clients -- obviously the server
    was unreachable for a short time and the client didn't notice.

    """

    def __init__(self, ui_display_callback,
                # pulseaudio pa_rtp
                pa_source, pa_sink, pa_source_volume, pa_sink_volume):
        """ The pa_* values are handed over to pa_rtp.PA_RTP. 
        pa_sink_volume is stored as default volume."""

        self.ui_display_callback = ui_display_callback

        # init audio streams (pa_rtp)
        self.pa_rtp = pa_rtp.PA_RTP(net_cfg.rtp_mcast_ip,
                                    pa_source,
                                    pa_sink,
                                    pa_source_volume,
                                    pa_sink_volume)
        self.pa_rtp.rx_mute(True)
        self.pa_rtp.tx_mute(True)
        self._vol_default = int(pa_sink_volume)
        self._vol = self._vol_default

        # state attributes
        self.state_main = 'idle'
        self.state_listen = False
        self.state_int = False

        # Lock and flags for thread-safe events
        self.ev_lock = RLock()

        # Timer
        self.timer = None
        self.timer_lock = RLock()

        # Server handling, thread safe
        self.server_lock = RLock()
        self._server_ip = None

        # create threads for network communication
        self.th_tcp_listen = Thread(target=self.tcp_listen, daemon=True)
        self.th_keep_connected = Thread(target=self.keep_connected, daemon=True)

    def start(self):
        # start threads for network communication
        self.th_tcp_listen.start()
        self.th_keep_connected.start()

    def done(self):
        self.timer_stop()
        self.send_PTT_OFF()
        # call done() for all instance attributes which implement it
        self.pa_rtp.done()

    # ------------------------------------------------------------------------
    # Event handling
    # ------------------------------------------------------------------------

    def update_rtp_audio(self):
        "Update RTP audio mute state for all event handlers."

        if self.state_main == 'idle':
            # mute speaker, mute mic
            self.pa_rtp.rx_mute(True)
            self.pa_rtp.tx_mute(True)
        elif self.state_main == 'recv':
            # unmute speaker, mute mic
            self.pa_rtp.rx_mute(False)
            self.pa_rtp.tx_mute(True)
        elif self.state_main == 'send':
            # mute speaker, unmute mic
            self.pa_rtp.rx_mute(True)
            self.pa_rtp.tx_mute(False)
        else:
            raise RuntimeError(f"_state_set(): '{self.state_main}'")
        self.pa_rtp.set_volume(self._vol)


    def ui_ev_PTT_press(self):
        with self.ev_lock:
            # never happens: pressing PTT while already in state 'send'
            if self.state_main == 'send': 
                raise RuntimeError("state_main is already 'send'.")
            # enter state 'send'
            if self.state_int:
                self.send_LISTEN_INT_REQ()
            else:
                self.send_PTT_ON()
                self.timer_ptt()
            # update state
            self.state_main = 'send'
            self.update_rtp_audio()
            self.ev_update_ui()

    def ui_ev_PTT_release(self):
        with self.ev_lock:
            # leaving state 'send'
            if not self.state_int:
                self.timer_stop()
                self.send_PTT_OFF()
            # entering state 'recv'
            if self.state_listen:
                self.timer_stop()
            else:
                self.timer_recv()
            # update state
            self.state_main = 'recv'
            self.update_rtp_audio()
            self.ev_update_ui()

    def ui_ev_press_int(self):
        with self.ev_lock:
            self.state_int = not self.state_int
            if self.state_int:
                # enter state internal
                self.send_LISTEN_INT_REQ()
                if self.state_main == 'send':
                    self.timer_stop()
                    self.send_PTT_OFF()
            else:
                # leave state internal
                if self.state_main == 'send':
                    self.send_PTT_ON()
                    self.timer_ptt()
            # update state
            self.ev_update_ui()

    def ui_ev_press_listen(self):
        with self.ev_lock:
            if self.state_main == 'idle':
                # listen requested
                self.state_main = 'recv'
                self.state_listen = True
            elif self.state_main == 'recv':
                # mute requested
                self.state_main = 'idle'
                self.state_listen = False
                self.timer_stop()
            else:
                # 'send'; just toggle listen state
                self.state_listen = not self.state_listen
            # update state
            self.update_rtp_audio()
            self.ev_update_ui()

    def ev_LISTEN(self):
        with self.ev_lock:
            if self.state_main == 'send' and not self.state_int:
                # This never should happen.
                print("ev_LISTEN: Got LISTEN while sending to extern.")
                # leaving state 'send'
                self.timer_off()
                self.send_PTT_OFF()
            # enter state 'recv'
            self.state_main = 'recv'
            self.state_int = False
            if self._vol < self._vol_default:
                self._vol = self._vol_default
            if not self.state_listen: self.timer_recv()
            # update state
            self.update_rtp_audio()
            self.ev_update_ui()

    def ev_LISTEN_INT(self):
        with self.ev_lock:
            if self.state_main == 'send' and not self.state_int:
                # This never should happen.
                print("ev_LISTEN_INT: LISTEN_INT while sending to extern.")
            elif self.state_main == 'send' and self.state_int:
                # We are sending internally already. Probably we started the
                # LISTEN-INT-REQ ourself. -> do nothing
                pass
            else:
                # enter state 'recv'
                self.state_main = 'recv'
                self.state_int = True
                if self._vol < self._vol_default:
                    self._vol = self._vol_default
                if not self.state_listen: self.timer_recv()
                # update state
                self.update_rtp_audio()
                self.ev_update_ui()

    def ev_LISTEN_INT_REJ(self):
        with self.ev_lock:
            if self.state_int:
                # We are in state_int although an external QSO is running
                # -> probably we sent LISTEN-INT-REQ
                # -> handle like LISTEN, so user can await termination of QSO.
                self.ev_LISTEN()

    def ev_recv_timeout(self):
        with self.ev_lock:
            self.state_main = 'idle'
            # update state
            self.update_rtp_audio()
            self.ev_update_ui()

    def ev_update_ui(self):
        with self.ev_lock:
            self.ui_display_callback(self.state_main == 'recv',
                                     self.state_main == 'send',
                                     self.state_listen,
                                     self.state_int,
                                     self.server_is_connected(),
                                     self._vol)

    # ------------------------------------------------------------------------
    # volume handling
    # ------------------------------------------------------------------------

    def ui_change_volume (self, d_vol):
        """Change volume by d_vol.
        The internal volume is in [0..100].
        The smallest volume step i +/- 1."""
        self._vol += d_vol
        if self._vol > 100: self._vol = 100
        if self._vol < 0: self._vol = 0
        self.pa_rtp.set_volume(self._vol)
        self.ev_update_ui()

    # ------------------------------------------------------------------------
    # networking: send events to server
    # ------------------------------------------------------------------------

    def send_PING(self):
        self._send_TCP(b'PING')

    def send_PTT_ON(self):
        self._send_TCP(b'PTT-ON')

    def send_PTT_OFF(self):
        self._send_TCP(b'PTT-OFF')

    def send_LISTEN_INT_REQ(self):
        self._send_TCP(b'LISTEN-INT-REQ')

    def _send_TCP(self, msg):
        if verbose: print(f'_send_TCP(): {msg}')
        try:
            if self.server_is_connected():
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    s.settimeout(1.0)
                    s.connect((self.server_get_ip(), net_cfg.tcp_server_port))
                    s.sendall(msg)
        except (ConnectionRefusedError, TimeoutError, OSError) as err:
            if verbose:
                print(f'_send_TCP(): error sending msg: {msg}')
                print(f'             {err}, {type(err)}')
            self.server_lost()

    # ------------------------------------------------------------------------
    # networking: look for server via UDP broadcast
    # ------------------------------------------------------------------------

    def keep_connected(self):
        """To be called as separate thread"""

        while True:
            if self.server_is_connected():
                self.send_PING()
                time.sleep(1.0)
            else:
                if verbose: print("keep_connected(): sending OVF-REGISTER")
                try:
                    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    s.settimeout(1.0)
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                    s.sendto(b'OVF-REGISTER',
                             ('<broadcast>', net_cfg.udp_server_port))
                except (ConnectionRefusedError, TimeoutError, OSError) as err:
                    if verbose:
                        print( 'keep_connected(): error sending OVF-REGISTER')
                        print(f'                  {err}, {type(err)}')
                time.sleep(0.1)
                if verbose: time.sleep(1)

    # ------------------------------------------------------------------------
    # networking: tcp listen
    # ------------------------------------------------------------------------

    def tcp_listen(self):
        """To be called as a separate thread"""

        # open TCP socket for listening
        # multiple connects are handled sequentially
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((net_cfg.tcp_client_bind_ip, net_cfg.tcp_client_port))
        s.listen(5) 

        # forever: handle incoming connections sequentially
        while True:
            c, addr = s.accept()
            c_ip = addr[0]              # address = ('ip-address', port<int>)
            with c:
                # Read all data until connection is closed.
                message = b''
                while True:
                    # read message, might come in multiple chunks
                    chunk = c.recv(1024)
                    if not chunk: break
                    message += chunk
                    # limit message to max length to avoid DoS attac
                    if len(message) > 255:
                        c.close()
                        break
                if verbose: print(f'tcp_listen(): read {message}')
                # Evaluate message
                if message == b'LISTEN':
                    self.ev_LISTEN()
                elif message == b'LISTEN-INT':
                    self.ev_LISTEN_INT()
                elif message == b'LISTEN-INT-REJ':
                    self.ev_LISTEN_INT_REJ()
                elif message == b'OVF-REGISTER-OK':
                    self.server_found(c_ip)

    # ------------------------------------------------------------------------
    # Timer
    # ------------------------------------------------------------------------
    # Actually two timers are involved: One for timer_recv and one for
    # timer_ptt. Combined start and stop functions are used to assure that
    # only one of them is running at once.

    def timer_stop(self):
        if verbose: print('timer_stop()')
        with self.timer_lock:
            if self.timer: self.timer.cancel()

    def timer_recv(self):
        if verbose: print('timer_recv() at ', datetime.datetime.now())
        with self.timer_lock:
            if self.timer: self.timer.cancel()
            self.timer = Timer(timeout_recv, self._timer_recv_expired)
            self.timer.start()

    def timer_ptt(self):
        if verbose: print('timer_ptt()')
        with self.timer_lock:
            if self.timer: self.timer.cancel()
            self.timer = Timer(time_ptt, self._timer_ptt_expired)
            self.timer.start()

    def _timer_ptt_expired(self):
        self.send_PTT_ON()
        self.timer_ptt()

    def _timer_recv_expired(self):
        if verbose: print('_timer_recv_expired() at ', 
                          datetime.datetime.now())
        self.ev_recv_timeout()

    # ------------------------------------------------------------------------
    # Server handling
    # ------------------------------------------------------------------------

    def server_is_connected(self):
        with self.server_lock:
            r = (self._server_ip != None)
        return r

    def server_lost(self):
        if verbose: print("server_lost()")
        with self.server_lock:
            self._server_ip = None
        self.ev_update_ui()

    def server_found(self, ip):
        if verbose: print(f"server_found(): {ip}")
        with self.server_lock:
            self._server_ip = ip
        self.pa_rtp.restart_rtp()
        self.ev_update_ui()

    def server_get_ip(self):
        with self.server_lock:
            r = self._server_ip
        return r

# vim: set et:
