#!/usr/bin/python3
from threading import Timer

verbose = False

# ----------------------------------------------------------------------------
# Factory function
# ----------------------------------------------------------------------------
# Import of needed modules
# import serial
# import gpiozero

def get_ptt(mode, port, ptt_hold_on=0.1):
    """Get PTT instance to control TRX.

    mode : string
        'rs232' or 'gpio'

    port : string or int
        example: '/dev/ttyUSB0' or 'GPIO17' or 17

    ptt_hold_on : float
        time in s to delay the PTT off(). This is to avoid truncating the
        emission too early due to digial audio delay.

    returns:
        instance of class PTT


    Usage
    -----

    my_ptt = get_ptt(mode, port, ptt_hold_on)
    my_ptt.on()
    my_ptt.off()
    my_ptt.done()
    """
    if mode == 'rs232':
        global serial
        import serial
        my_ptt = PTT_RS232(port, ptt_hold_on)
        return  my_ptt
    elif mode == 'gpio':
        global gpiozero 
        import gpiozero 
        my_ptt = PTT_GPIO(port, ptt_hold_on)
        return  my_ptt
    else:
        raise ValueError(f'unknown type "{mode}"')

# ----------------------------------------------------------------------------
# PTT interface
# ----------------------------------------------------------------------------

class PTT:
    """"Abstract class defining the PTT interface.

    Implementation
    --------------

    _open():        setting up the hardware
    _close():       restoring hardware, clean up
    _ptt_on():      swichting the ptt pyhsically on
    _ptt_off():     swichting the ptt pyhsically off

    """

    def __init__(self, port, ptt_hold_on):
        if verbose: print('ptt_init()')
        self.ptt_hold_on = ptt_hold_on
        self.timer = None
        self._open(port)

    def done(self):
        # assure PTT is released
        if verbose: print('ptt_done()')
        self._ptt_off()
        self._close()

    def on(self): 
        if verbose: print('ptt_on()')
        if self.timer: self.timer.cancel()
        self._ptt_on()

    def off(self): 
        if verbose: print('ptt_off() ...')
        if self.timer: self.timer.cancel()
        self.timer = Timer(self.ptt_hold_on, self._timer_expired)
        self.timer.start()

    def _timer_expired(self):
        if verbose: print('ptt_off()')
        self._ptt_off()

    def _hw_set_up(self, port):
        raise NotImplementedError()

    def _ptt_on(self):
        raise NotImplementedError()

    def _ptt_off(self):
        raise NotImplementedError()

    def _done(self):
        raise NotImplementedError()

# ----------------------------------------------------------------------------
# RS232 implementation
# ----------------------------------------------------------------------------

class PTT_RS232(PTT):

    def _open(self, port):
        self.ptt_serial = serial.Serial(port)
        self.ptt_serial.rts = False

    def _close(self):
        self.ptt_serial.close()

    def _ptt_on(self):
        self.ptt_serial.rts = True

    def _ptt_off(self):
        self.ptt_serial.rts = False

# ----------------------------------------------------------------------------
# GPIO implementation
# ----------------------------------------------------------------------------

class PTT_GPIO(PTT):
    r"""Connection of PTT
                                       ,------ PTT TRX
                                       |
                   ,------,          |/
        GPIO  o----| 10 k |--------->|
        3.3 V      '------'          |\E
                                       |
        GPIO = hi -> PTT on            |
        GPIO = lo -> PTT off          ===
    """

    def _open(self, port):
        self.pin = gpiozero.DigitalOutputDevice(port)

    def _close(self):
        self.pin.off()

    def _ptt_on(self):
        self.pin.on()

    def _ptt_off(self):
        self.pin.off()


# vim: set et fdm=marker:
